;Recursive Quicksort by bitRAKE

;
;;Initial State: (general)
;;
;;   esi      ecx      edi
;;    |---n---|-|---n---|
;;             P
;;
;;My choice of Pivot:
;;
;;      esi             edi
;;    |-|-------n-------|
;;     P
;;
;;
;;Possible Results: (P = pivot index)
;;
;;      esi             edi
;;    |-|-------n-------|     all items are >= pivot
;;     P                      inc esi, partition new range
;;
;;        esi           edi
;;    |-|-|-----n-------|     one item =< pivot, rest >= pivot
;;       P                    inc esi, inc esi, partition new range
;;
;;  esi--n--edi esi--n--edi
;;    |---n---|-|---n---|     some items =< pivot, some items >= pivot
;;             P              partition range #1, partition range #1
;;
;;  esi           edi
;;    |------n------|-|-|     one item >= pivot, rest <= pivot
;;                   P        dec edi, partition new range
;;
;;  esi             edi
;;    |------n--------|-|     all items are <= pivot
;;                     P      dec edi, dec edi, partition new range
;;

.data

lpQMem		dd ?
lpQStr		dd ?
fQDesc		dd ?

.code

QuickSort PROC uses esi edi ebx,qARRAY:DWORD,qLOW:DWORD,qHIGH:DWORD,lpMem:DWORD,lpStr:DWORD,fDesc:DWORD
	mov		eax,lpMem
	mov		lpQMem,eax
	mov		eax,lpStr
	mov		lpQStr,eax
	mov		eax,fDesc
	mov		fQDesc,eax
	push	ebp
	mov		ebx,qARRAY
	mov		esi,qLOW
	mov		edi,qHIGH
	call	PARTITION
	pop		ebp
	ret

PARTITION:
	push	esi
	push	edi
	mov		ebp,edi
	sub		edi,esi		; width of partition - 2
	jle		@exit		; must be >1 elements to sort
	; bad choice for sorted data
	mov		ecx,esi		; choose pivot
	; better choice?
;	lea		ecx,[esi+ebp]
;	shr		ecx,1
;Note: You can choose any pivot you want - algorithm doesn't move
;  items that equal the pivot - including the pivot, which is
;  required!  A general solution would choose the mid-point item
;  for the case of partially sorted data.  I have chosen the first
;  item to reduce the size of the partition at the start.  Uncomment
;  the following line if you choose another pivot method - it is not
;  needed because pivot item lies between partitions by definition.
;	dec		esi			; counter first inc
	inc		ebp			; counter first dec
  @low:
	inc		esi
	dec		edi
	js		@donel ; esi is out of range or part of upper partition
	.if lpQStr
		mov		edx,[ebx+esi*4]
		call	@CompareStr
	.else
		mov		eax,lpQMem
		mov		edx,eax
		add		eax,[ebx+ecx*4]
		add		edx,[ebx+esi*4]
		mov		eax,[eax]
		sub		eax,[edx]
	.endif
	.if fQDesc
		neg		eax
	.endif
	cmp		eax,0
	jge		@low
  @high:
	dec		ebp
	dec		edi
	js		@donel ; ebp is out of range or part of lower partition
	.if lpQStr
		mov		edx,[ebx+ebp*4]
		call	@CompareStr
	.else
		mov		eax,lpQMem
		mov		edx,eax
		add		eax,[ebx+ecx*4]
		add		edx,[ebx+ebp*4]
		mov		eax,[eax]
		sub		eax,[edx]
	.endif
	.if fQDesc
		neg		eax
	.endif
	cmp		eax,0
	jle		@high
	mov		edx,[ebx+esi*4]
	mov		eax,[ebx+ebp*4]
	mov		[ebx+ebp*4],edx
	mov		[ebx+esi*4],eax
	jmp		@low
  @donel:
	dec		esi
	mov		edx,[ebx+ecx*4]
	mov		eax,[ebx+esi*4]
	mov		[ebx+esi*4],edx
	mov		[ebx+ecx*4],eax
	dec		esi ; pivot is sorted
	mov		edi,[esp]
	mov		[esp],esi
	mov		esi,ebp
	call	PARTITION
	pop		edi
	pop		esi
	jmp		PARTITION
  @exit:
	add		esp,8
	retn

@CompareStr:
	push	edi
	mov		edi,[ebx+ecx*4]
	mov		eax,lpQMem
	lea		edi,[edi+eax]
	lea		edx,[edx+eax]
	mov		edi,[edi]
	mov		edx,[edx]
	mov		eax,lpQStr
	lea		edi,[edi+eax]
	lea		edx,[edx+eax]
	jmp		@f
  @next:
	inc		edi
	inc		edx
  @@:
	mov		al,[edi]
	mov		ah,[edx]
	or		ax,ax
	je		@exitcmpstr
	cmp		al,'a'
	jb		@f
	cmp		al,'z'
	jg		@f
	and		al,5Fh
  @@:
	cmp		ah,'a'
	jb		@f
	cmp		ah,'z'
	jg		@f
	and		ah,5Fh
  @@:
	sub		al,ah
	je		@next
  @exitcmpstr:
	cbw
	cwde
	pop		edi
	retn

QuickSort ENDP
